"""
These functions are used to identify areas that meet a number of different criteria you specify. These criteria can be based
upon attribute queries (for example, parcels that are vacant) and spatial queries (for example, within 1 kilometer of a
river). The areas that are found can be selected from existing features (such as existing land parcels) or new features
can be created where all the requirements are met.

find_existing_locations searches for existing areas in a layer that meet a series of criteria.
derive_new_locations creates new areas from locations in your study area that meet a series of criteria.
find_similar_locations finds locations most similar to one or more reference locations based on criteria you specify.
find_centroids finds and generates points from the representative center (centroid) of each input multipoint, line, or area feature.
choose_best_facilities choose the best locations for facilities by allocating locations that have demand for these
facilities in a way that satisfies a given goal.
create_viewshed creates areas that are visible based on locations you specify.
create_watersheds creates catchment areas based on locations you specify.
trace_downstream determines the flow paths in a downstream direction from the locations you specify
"""
import arcgis as _arcgis

import arcgis.network as network


def find_existing_locations(
        input_layers=None,
        expressions=None,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/find_existing_locations/find_existing_locations.png

    .. |intersect| image:: _static/images/derive_new_locations/intersect.png
    .. |distance| image:: _static/images/derive_new_locations/distance.png
    .. |within| image:: _static/images/derive_new_locations/within.png
    .. |nearest| image:: _static/images/derive_new_locations/nearest.png
    .. |contains| image:: _static/images/derive_new_locations/contains.png

    The ``find_existing_locations`` method selects features in the input layer that meet a query you specify.
    A query is made up of one or more expressions. There are two types of expressions: attribute and spatial.
    An example of an attribute expression is that a parcel must be vacant, which is an attribute of the Parcels layer (where STATUS = 'VACANT').
    An example of a spatial expression is that the parcel must also be within a certain distance of a river (Parcels within a distance of 0.75 Miles from Rivers).

    =====================================    ======================================================================================================
    **Argument**                             **Description**
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    input_layers                             Required list of feature layers. A list of layers that will be used in the expressions parameter.
                                             Each layer in the list can be:

                                             * a feature service layer with an optional filter to select specific features, or
                                             * a feature collection

                                             See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    expressions                              Required dict. There are two types of expressions, attribute and spatial.

                                             Example attribute expression:

                                             {
                                                "operator": "and",
                                                "layer": 0,
                                                "where": "STATUS = 'VACANT'"
                                             }

                                             **Note**

                                             * operator can be either ``and`` or ``or``
                                             * layer is the index of the layer in the ``input_layers`` parameter.
                                             * The where clause must be surrounded by double quotes.
                                             * When dealing with text fields, values must be single-quoted ('VACANT').
                                             * Date fields support all queries except LIKE. Dates are strings in YYYY:MM:DD hh:mm:ss format.
                                              Here's an example using the date field ObsDate:

                                              "where": "ObsDate >= '1998-04-30 13:30:00' "

                                             +----------+------------------------------------------------------------------+
                                             | =        | Equal                                                            |
                                             +----------+------------------------------------------------------------------+
                                             | >        | Greater than                                                     |
                                             +----------+------------------------------------------------------------------+
                                             | <        | Less than                                                        |
                                             +----------+------------------------------------------------------------------+
                                             | >=       | Greater than or equal to                                         |
                                             +----------+------------------------------------------------------------------+
                                             | <=       | Less than or equal to                                            |
                                             +----------+------------------------------------------------------------------+
                                             | <>       | Not equal                                                        |
                                             +----------+------------------------------------------------------------------+
                                             | LIKE '%  | A percent symbol (%) signifies a wildcard, meaning that          |
                                             | <string>'| anything is acceptable in its place-one character, a             |
                                             |          | hundred characters, or no character. This expression             |
                                             |          | would select Mississippi and Missouri among USA                  |
                                             |          | state names: STATE_NAME LIKE 'Miss%'                             |
                                             +----------+------------------------------------------------------------------+
                                             | BETWEEN  | Selects a record if it has a value greater than or equal         |
                                             | <value1> | to <value1> and less than or equal to <value2>.                  |
                                             | AND      | For example, this expression selects all records with            |
                                             | <value2> | an HHSIZE value greater than or equal to 3 and less              |
                                             |          | than or equal to 10:                                             |
                                             |          |                                                                  |
                                             |          | HHSIZE BETWEEN 3 AND 10                                          |
                                             |          |                                                                  |
                                             |          | The above is equivalent to:                                      |
                                             |          |                                                                  |
                                             |          | HHSIZE >= 3 AND HHSIZE <= 10                                     |
                                             |          | This operator applies to numeric or date fields.                 |
                                             |          | Here is an example of a date query on the field ObsDate:         |
                                             |          |                                                                  |
                                             |          | ObsDate BETWEEN '1998-04-30 00:00:00' AND '1998-04-30 23:59:59'  |
                                             |          |                                                                  |
                                             |          | Time is optional.                                                |
                                             +----------+------------------------------------------------------------------+
                                             | NOT      | Selects a record if it has a value outside the range between     |
                                             | BETWEEN  | <value1> and less than or equal to <value2>.                     |
                                             | <value1> | For example, this expression selects all records whose           |
                                             | AND      | HHSIZE value is less than 5 and greater than 7.                  |
                                             | <value2> |                                                                  |
                                             |          | HHSIZE NOT BETWEEN 5 AND 7                                       |
                                             |          |                                                                  |
                                             |          | The above is equivalent to:                                      |
                                             |          |                                                                  |
                                             |          | HHSIZE < 5 OR HHSIZE > 7                                         |
                                             |          | This operator applies to numeric or date fields.                 |
                                             |          |                                                                  |
                                             |          | **Note**                                                         |
                                             |          |                                                                  |
                                             |          | You can use the contains relationship with points and lines.     |
                                             |          | For example, you have a layer of street centerlines (lines) and  |
                                             |          | a layer of manhole covers (points), and you want to find streets |
                                             |          | that contain a manhole cover. You could use contains to find     |
                                             |          | streets that contain manhole covers, but in order for a line to  |
                                             |          | contain a point, the point must be exactly on the line (that is, |
                                             |          | in GIS terms, they are snapped to each other). If there is any   |
                                             |          | doubt about this, use the withinDistance relationship with a     |
                                             |          | suitable distance value.                                         |
                                             +----------+------------------------------------------------------------------+

                                             Example spatial expression:
                                             {
                                                "operator": "and",
                                                "layer": 0,
                                                "spatialRel": "withinDistance",
                                                "selectingLayer": 1,
                                                "distance": 10,
                                                "units": "miles"
                                             }

                                             * operator can be either ``and`` or ``or``
                                             * layer is the index of the layer in ``the input_layers`` parameter. The result of the expression is features in this layer.
                                             * spatialRel is the spatial relationship. There are nine spatial relationships.
                                             * distance is the distance to use for the withinDistance and notWithinDistance spatial relationship.
                                             * units is the units for distance.

                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | spatialRel        | Description                                                                            |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | intersects        | |intersect|                                                                            |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes the intersect test if it overlaps                            |
                                             | notIntersects     | any part of a feature in selectingLayer, including touches                             |
                                             |                   | (where features share a common point).                                                 |
                                             |                   |                                                                                        |
                                             |                   | * intersects-If a feature in layer intersects a feature in                             |
                                             |                   |   selectingLayer, the portion of the feature in layer that                             |
                                             |                   |   intersects the feature in selectingLayer is included in                              |
                                             |                   |   the output.                                                                          |
                                             |                   | * notintersects-If a feature in layer intersects a feature in                          |
                                             |                   |   selectingLayer, the portion of the feature in layer that                             |
                                             |                   |   intersects the feature in selectingLayer is excluded from                            |
                                             |                   |   the output.                                                                          |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | withinDistance    | |distance|                                                                             |
                                             |                   |                                                                                        |
                                             |                   | The within a distance relationship uses the straight-line                              |
                                             | notWithinDistance | distance between features in layer to those in selectingLayer.                         |
                                             |                   | withinDistance-The portion of the feature in layer that is                             |
                                             |                   | within the specified distance of a feature in selectingLayer                           |
                                             |                   | is included in the output.                                                             |
                                             |                   | notwithinDistance-The portion of the feature in layer that is                          |
                                             |                   | within the specified distance of a feature in selectingLayer is                        |
                                             |                   | excluded from output. You can think of this relationship as                            |
                                             |                   | "is farther away than".                                                                |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | contains          | |intersect|                                                                            |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes this test if it completely                                   |
                                             | notContains       | surrounds a feature in selectingLayer. No portion of the                               |
                                             |                   | containing feature; however, the contained feature is allowed                          |
                                             |                   | to touch the containing feature (that is, share a common                               |
                                             |                   | point along its boundary).                                                             |
                                             |                   |                                                                                        |
                                             |                   | contains-If a feature in layer contains a feature in                                   |
                                             |                   | selectingLayer, the feature in layer is included in the output.                        |
                                             |                   | notcontains-If a feature in layer contains a feature in                                |
                                             |                   | selectingLayer, the feature in the first layer is excluded                             |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | within            | |within|                                                                               |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes this test if it is completely                                |
                                             | notWithin         | surrounded by a feature in selectingLayer. The entire feature                          |
                                             |                   | layer must be within the containing feature; however, the two                          |
                                             |                   | features are allowed to touch (that is, share a common point                           |
                                             |                   | along its boundary).                                                                   |
                                             |                   |                                                                                        |
                                             |                   | * within-If a feature in layer is completely within a feature in                       |
                                             |                   |   selectingLayer, the feature in layer is included in the output.                      |
                                             |                   | * notwithin-If a feature in layer is completely within a feature                       |
                                             |                   |   in selectingLayer, the feature in layer is excluded from the                         |
                                             |                   |   output.                                                                              |
                                             |                   |                                                                                        |
                                             |                   | **Note:**                                                                              |
                                             |                   |                                                                                        |
                                             |                   | can use the within relationship for points and lines, just as                          |
                                             |                   | you can with the contains relationship. For example, your first                        |
                                             |                   | layer contains points representing manhole covers and you want                         |
                                             |                   | to find the manholes that are on street centerlines (as opposed                        |
                                             |                   | to parking lots or other non-street features). You could use                           |
                                             |                   | within to find manhole points within street centerlines, but                           |
                                             |                   | in order for a point to contain a line, the point must be exactly                      |
                                             |                   | on the line (that is, in GIS terms, they are snapped to each                           |
                                             |                   | other). If there is any doubt about this, use the withinDistance                       |
                                             |                   | relationship with a suitable distance value.                                           |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | nearest           | |nearest|                                                                              |
                                             |                   |                                                                                        |
                                             |                   | feature in the first layer passes this test if it is nearest                           |
                                             |                   | to a feature in the second layer.                                                      |
                                             |                   |                                                                                        |
                                             |                   | * nearest-If a feature in the first layer is nearest to a                              |
                                             |                   |   feature in the second layer, the feature in the first layer                          |
                                             |                   |   is included in the output.                                                           |
                                             +-------------------+----------------------------------------------------------------------------------------+

                                             * ``distance`` is the distance to use for the withinDistance and notWithinDistance spatial relationship.
                                             * ``units`` is the units for distance.

                                             Choice list: ['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles']

                                             An expression may be a list, which denotes a group. The first operator in the group indicates how the group expression
                                             is added to the previous expression. Grouping expressions is only necessary when you need to create two or more distinct
                                             sets of features from the same layer. One way to think of grouping is that without grouping, you would have to execute
                                             ``find_existing_locations`` multiple times and merge the results.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    output_name                              Optional string. If provided, the method will create a feature layer of the results. You define the name of the layer.
                                             If ``output_name`` is not supplied, the task will return a feature collection.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    context                                  Optional string. Additional settings such as processing extent and output spatial reference. For ``find_existing_locations``, there are two settings.

                                             #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the input_layers that intersect the bounding box will be analyzed.
                                             #. Output Spatial Reference (``outSR``)-the output features will be projected into the output spatial reference.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    gis                                      Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    estimate                                 Optional boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================    ======================================================================================================

    :Returns: result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        #USAGE EXAMPLE: To find busy (where SEGMENT_TY is 1 and where ARTERIAL_C is 1) streets from the existing seattle streets layer.

        arterial_streets = find_existing_locations(input_layers=[bike_route_streets],
                        expressions=[{"operator":"","layer":0,"where":"SEGMENT_TY = 1"},
                                     {"operator":"and","layer":0,"where":"ARTERIAL_C = 1"}],
                                       output_name='ArterialStreets')



    """

    if input_layers is None:
        input_layers = []
    if expressions is None:
        expressions = []
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.find_existing_locations(
        input_layers,
        expressions,
        output_name,
        context,
        estimate=estimate,future=future)


def derive_new_locations(
        input_layers=[],
        expressions=[],
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/derive_new_locations/derive_new_locations.png

    .. |intersect| image:: _static/images/derive_new_locations/intersect.png
    .. |distance| image:: _static/images/derive_new_locations/distance.png
    .. |within| image:: _static/images/derive_new_locations/within.png
    .. |nearest| image:: _static/images/derive_new_locations/nearest.png
    .. |contains| image:: _static/images/derive_new_locations/contains.png

    The ``derive_new_locations`` method derives new features from the input layers that meet a query you specify. A query is
    made up of one or more expressions. There are two types of expressions: attribute and spatial. An example of an
    attribute expression is that a parcel must be vacant, which is an attribute of the Parcels layer
    (STATUS = 'VACANT'). An example of a spatial expression is that the parcel must also be within a certain
    distance of a river (Parcels within a distance of 0.75 Miles from Rivers).

    The ``derive_new_locations`` method is very similar to the ``find_existing_locations`` method, the main difference is that
    the result of ``derive_new_locations`` can contain partial features.

    * In both methods, the attribute expression  ``where`` and the spatial relationships within and contains return the same result.
      This is because these relationships return entire features.
    * When ``intersects`` or ``within_distance`` is used, ``derive_new_locations`` creates new features
      in the result. For example, when intersecting a parcel feature and a flood zone area that partially overlap each other,
      ``find_existing_locations`` will return the entire parcel whereas ``derive_new_locations`` will return just the portion of
      the parcel that is within the flood zone.

    =====================================    ======================================================================================================
    **Argument**                             **Description**
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    input_layers                             Required list of feature layers. A list of layers that will be used in the expressions parameter.
                                             Each layer in the list can be:

                                             * a feature service layer with an optional filter to select specific features, or
                                             * a feature collection
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    expressions                              Required dict. There are two types of expressions, attribute and spatial.

                                             Example attribute expression:

                                             {
                                                "operator": "and",
                                                "layer": 0,
                                                "where": "STATUS = 'VACANT'"
                                             }

                                             **Note**
                                             * operator can be either ``and`` or ``or``
                                             * layer is the index of the layer in the ``input_layers`` parameter.
                                             * The where clause must be surrounded by double quotes.
                                             * When dealing with text fields, values must be single-quoted ('VACANT').
                                             * Date fields support all queries except LIKE. Dates are strings in YYYY:MM:DD hh:mm:ss format. Here's an example using the date field ObsDate:
                                               "where": "ObsDate >= '1998-04-30 13:30:00' "

                                             +----------+------------------------------------------------------------------+
                                             | =        | Equal                                                            |
                                             +----------+------------------------------------------------------------------+
                                             | >        | Greater than                                                     |
                                             +----------+------------------------------------------------------------------+
                                             | <        | Less than                                                        |
                                             +----------+------------------------------------------------------------------+
                                             | >=       | Greater than or equal to                                         |
                                             +----------+------------------------------------------------------------------+
                                             | <=       | Less than or equal to                                            |
                                             +----------+------------------------------------------------------------------+
                                             | <>       | Not equal                                                        |
                                             +----------+------------------------------------------------------------------+
                                             | LIKE '%  | A percent symbol (%) signifies a wildcard, meaning that          |
                                             | <string>'| anything is acceptable in its place-one character, a             |
                                             |          | hundred characters, or no character. This expression             |
                                             |          | would select Mississippi and Missouri among USA                  |
                                             |          | state names: STATE_NAME LIKE 'Miss%'                             |
                                             +----------+------------------------------------------------------------------+
                                             | BETWEEN  | Selects a record if it has a value greater than or equal         |
                                             | <value1> | to <value1> and less than or equal to <value2>.                  |
                                             | AND      | For example, this expression selects all records with            |
                                             | <value2> | an HHSIZE value greater than or equal to 3 and less              |
                                             |          | than or equal to 10:                                             |
                                             |          |                                                                  |
                                             |          | HHSIZE BETWEEN 3 AND 10                                          |
                                             |          |                                                                  |
                                             |          | The above is equivalent to:                                      |
                                             |          |                                                                  |
                                             |          | HHSIZE >= 3 AND HHSIZE <= 10                                     |
                                             |          | This operator applies to numeric or date fields.                 |
                                             |          | Here is an example of a date query on the field ObsDate:         |
                                             |          |                                                                  |
                                             |          | ObsDate BETWEEN '1998-04-30 00:00:00' AND '1998-04-30 23:59:59'  |
                                             |          |                                                                  |
                                             |          | Time is optional.                                                |
                                             +----------+------------------------------------------------------------------+
                                             | NOT      | Selects a record if it has a value outside the range between     |
                                             | BETWEEN  | <value1> and less than or equal to <value2>.                     |
                                             | <value1> | For example, this expression selects all records whose           |
                                             | AND      | HHSIZE value is less than 5 and greater than 7.                  |
                                             | <value2> |                                                                  |
                                             |          | HHSIZE NOT BETWEEN 5 AND 7                                       |
                                             |          |                                                                  |
                                             |          | The above is equivalent to:                                      |
                                             |          |                                                                  |
                                             |          | HHSIZE < 5 OR HHSIZE > 7                                         |
                                             |          | This operator applies to numeric or date fields.                 |
                                             |          |                                                                  |
                                             |          | **Note**                                                         |
                                             |          |                                                                  |
                                             |          | You can use the contains relationship with points and lines.     |
                                             |          | For example, you have a layer of street centerlines (lines) and  |
                                             |          | a layer of manhole covers (points), and you want to find streets |
                                             |          | that contain a manhole cover. You could use contains to find     |
                                             |          | streets that contain manhole covers, but in order for a line to  |
                                             |          | contain a point, the point must be exactly on the line (that is, |
                                             |          | in GIS terms, they are snapped to each other). If there is any   |
                                             |          | doubt about this, use the withinDistance relationship with a     |
                                             |          | suitable distance value.                                         |
                                             +----------+------------------------------------------------------------------+

                                             Example spatial expression:
                                             {
                                                "operator": "and",
                                                "layer": 0,
                                                "spatialRel": "withinDistance",
                                                "selectingLayer": 1,
                                                "distance": 10,
                                                "units": "miles"
                                             }

                                             * operator can be either ``and`` or ``or``
                                             * layer is the index of the layer in ``the input_layers`` parameter. The result of the expression is features in this layer.
                                             * spatialRel is the spatial relationship. There are nine spatial relationships.
                                             * distance is the distance to use for the withinDistance and notWithinDistance spatial relationship.
                                             * units is the units for distance.

                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | spatialRel        | Description                                                                            |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | intersects        | |intersect|                                                                            |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes the intersect test if it overlaps                            |
                                             | notIntersects     | any part of a feature in selectingLayer, including touches                             |
                                             |                   | (where features share a common point).                                                 |
                                             |                   |                                                                                        |
                                             |                   | * intersects-If a feature in layer intersects a feature in                             |
                                             |                   |   selectingLayer, the portion of the feature in layer that                             |
                                             |                   |   intersects the feature in selectingLayer is included in                              |
                                             |                   |   the output.                                                                          |
                                             |                   | * notintersects-If a feature in layer intersects a feature in                          |
                                             |                   |   selectingLayer, the portion of the feature in layer that                             |
                                             |                   |   intersects the feature in selectingLayer is excluded from                            |
                                             |                   |   the output.                                                                          |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | withinDistance    | |distance|                                                                             |
                                             |                   |                                                                                        |
                                             |                   | The within a distance relationship uses the straight-line                              |
                                             | notWithinDistance | distance between features in layer to those in selectingLayer.                         |
                                             |                   | withinDistance-The portion of the feature in layer that is                             |
                                             |                   | within the specified distance of a feature in selectingLayer                           |
                                             |                   | is included in the output.                                                             |
                                             |                   | notwithinDistance-The portion of the feature in layer that is                          |
                                             |                   | within the specified distance of a feature in selectingLayer is                        |
                                             |                   | excluded from output. You can think of this relationship as                            |
                                             |                   | "is farther away than".                                                                |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | contains          | |intersect|                                                                            |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes this test if it completely                                   |
                                             | notContains       | surrounds a feature in selectingLayer. No portion of the                               |
                                             |                   | containing feature; however, the contained feature is allowed                          |
                                             |                   | to touch the containing feature (that is, share a common                               |
                                             |                   | point along its boundary).                                                             |
                                             |                   |                                                                                        |
                                             |                   | contains-If a feature in layer contains a feature in                                   |
                                             |                   | selectingLayer, the feature in layer is included in the output.                        |
                                             |                   | notcontains-If a feature in layer contains a feature in                                |
                                             |                   | selectingLayer, the feature in the first layer is excluded                             |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | within            | |within|                                                                               |
                                             |                   |                                                                                        |
                                             |                   | A feature in layer passes this test if it is completely                                |
                                             | notWithin         | surrounded by a feature in selectingLayer. The entire feature                          |
                                             |                   | layer must be within the containing feature; however, the two                          |
                                             |                   | features are allowed to touch (that is, share a common point                           |
                                             |                   | along its boundary).                                                                   |
                                             |                   |                                                                                        |
                                             |                   | * within-If a feature in layer is completely within a feature in                       |
                                             |                   |   selectingLayer, the feature in layer is included in the output.                      |
                                             |                   | * notwithin-If a feature in layer is completely within a feature                       |
                                             |                   |   in selectingLayer, the feature in layer is excluded from the                         |
                                             |                   |   output.                                                                              |
                                             |                   |                                                                                        |
                                             |                   | **Note:**                                                                              |
                                             |                   |                                                                                        |
                                             |                   | can use the within relationship for points and lines, just as                          |
                                             |                   | you can with the contains relationship. For example, your first                        |
                                             |                   | layer contains points representing manhole covers and you want                         |
                                             |                   | to find the manholes that are on street centerlines (as opposed                        |
                                             |                   | to parking lots or other non-street features). You could use                           |
                                             |                   | within to find manhole points within street centerlines, but                           |
                                             |                   | in order for a point to contain a line, the point must be exactly                      |
                                             |                   | on the line (that is, in GIS terms, they are snapped to each                           |
                                             |                   | other). If there is any doubt about this, use the withinDistance                       |
                                             |                   | relationship with a suitable distance value.                                           |
                                             +-------------------+----------------------------------------------------------------------------------------+
                                             | nearest           | |nearest|                                                                              |
                                             |                   |                                                                                        |
                                             |                   | feature in the first layer passes this test if it is nearest                           |
                                             |                   | to a feature in the second layer.                                                      |
                                             |                   |                                                                                        |
                                             |                   | * nearest-If a feature in the first layer is nearest to a                              |
                                             |                   |   feature in the second layer, the feature in the first layer                          |
                                             |                   |   is included in the output.                                                           |
                                             +-------------------+----------------------------------------------------------------------------------------+

    -------------------------------------    ------------------------------------------------------------------------------------------------------
    output_name                              Optional string. If provided, the task will create a feature layer of the results. You define the name of the layer.                                           If output_name is not supplied, the task will return a feature collection.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    context                                  Optional string. Additional settings such as processing extent and output spatial reference. For                                                               ``derive_new_locations``, there are two settings.

                                             #. Extent (extent)-a bounding box that defines the analysis area. Only those points in the input_layers that intersect the bounding box will be analyzed.
                                             #. Output Spatial Reference (outSR)
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    gis                                      Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    estimate                                 Optional boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================    ======================================================================================================

    .. code-block:: python

        USAGE EXAMPLE: To Identify areas that are suitable cougar habitat using the criteria defined by experts.

        new_loaction = derive_new_locations(input_layers=[slope, vegetation, streams, highways],
                                    expressions=[{"operator":"","layer":0,"selectingLayer":1,"spatialRel":"intersects"},
                                                 {"operator":"and","layer":0,"selectingLayer":2,"spatialRel":"withinDistance","distance":500,"units":"Feet"},
                                                 {"operator":"and","layer":0,"selectingLayer":3,"spatialRel":"notWithinDistance","distance":1500,"units":"Feet"},
                                                 {"operator":"and","layer":0,"where":"GRIDCODE = 1"}],
                                    output_name='derive_new_loactions')



    """

    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.derive_new_locations(
        input_layers,
        expressions,
        output_name,
        context,
        estimate=estimate,
        future=future)

def find_similar_locations(
        input_layer,
        search_layer,
        analysis_fields=[],
        input_query=None,
        number_of_results=0,
        output_name=None,
        context=None,
        gis=None, estimate=False, future=False):
    """
    .. image:: _static/images/find_similar_locations/find_similar_locations.png

    The ``find_similar_locations`` method measures the similarity of candidate locations to one or more reference locations.

    Based on criteria you specify, Find ``find_similar_locations`` can answer questions such as the following:

    Which of your stores are most similar to your top performers with regard to customer profiles?
    Based on characteristics of villages hardest hit by the disease, which other villages are high risk?
    To answer questions such as these, you provide the reference locations (the ``input_layer`` parameter), the candidate
    locations (the ``search_layer`` parameter), and the fields representing the criteria you want to match. For example,
    the ``input_layer`` might be a layer containing your top performing stores or the villages hardest hit by the disease.
    The ``search_layer`` contains your candidate locations to search. This might be all of your stores or all other villages.
    Finally, you supply a list of fields to use for measuring similarity. ``find_similar_locations`` will rank all of the
    candidate locations by how closely they match your reference locations across all of the fields you have selected.

    =======================  ===========================================================================================
    **Argument**             **Description**
    -----------------------  -------------------------------------------------------------------------------------------
    input_layer              Required feature layer. The ``input_layer`` contains one or more
                             reference locations against which features in the ``search_layer``
                             will be evaluated for similarity. For example, the ``input_layer``
                             might contain your top performing stores or the villages hardest
                             hit by a disease.
                             It is not uncommon that the ``input_layer`` and ``search_layer`` are the
                             same feature service. For example, the feature service contains
                             locations of all stores, one of which is your top performing store.
                             If you want to rank the remaining stores from most to least similar
                             to your top performing store, you can provide a filter for both the
                             inputLayer and the ``search_layer``. The filter on the ``input_layer`` would
                             select the top performing store while the filter on the ``search_layer``
                             would select all stores except for the top performing store. You can
                             also use the optional ``input_query`` parameter to specify reference locations.

                             If there is more than one reference location, similarity will be based
                             on averages for the fields you specify in the ``analysis_fields`` parameter.
                             So, for example, if there are two reference locations and you are
                             interested in matching population, the task will look for candidate
                             locations in the ``search_layer`` with populations that are most like the
                             average population for both reference locations. If the values for the
                             reference locations are 100 and 102, for example, the method will look
                             for candidate locations with populations near 101. Consequently, you
                             will want to use fields for the reference locations fields that have
                             similar values. If, for example, the population values for one reference
                             location is 100 and the other is 100,000, the tool will look for candidate
                             locations with population values near the average of those two values: 50,050.
                             Notice that this averaged value is nothing like the population for either
                             of the reference locations. See :ref:`Feature Input<FeatureInput>`.
    -----------------------  -------------------------------------------------------------------------------------------
    search_layer             Required feature layer. The layer containing candidate locations that
                             will be evaluated against the reference locations. See :ref:`Feature Input<FeatureInput>`.
    -----------------------  -------------------------------------------------------------------------------------------
    analysis_fields          Required list of strings. A list of fields whose values are used to determine similarity.
                             They must be numeric fields and the fields must exist on both the ``input_layer`` and
                             the ``search_layer``. The method will find features in the ``search_layer`` that have field
                             values closest to those of the features in your ``input_layer``.
    -----------------------  -------------------------------------------------------------------------------------------
    input_query              Optional string. In the situation where the ``input_layer`` and the ``search_layer`` are the same feature service,
                             this parameter allows you to input a query on the ``input_layer`` to specify which features are the reference locations.
                             The reference locations specified by this query will not be analyzed as candidates.
                             The syntax of ``input_query`` is the same as a filter.
    -----------------------  -------------------------------------------------------------------------------------------
    number_of_results        Optional int. The number of ranked candidate locations output to the ``similar_result_layer``.
                             If ``number_of_results`` is not specified, or set to zero, all candidate locations will be ranked and output.
    -----------------------  -------------------------------------------------------------------------------------------
    output_name              Optional string. If provided, the method will create a feature service of the results.
                             You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    -----------------------  -------------------------------------------------------------------------------------------
    context                  Optional string. Context contains additional settings that affect method execution.
                             For ``find_similar_locations``, there are two settings.

                             #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features
                                in the ``input_layer`` that intersect the bounding box will be analyzed.
                             #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -----------------------  -------------------------------------------------------------------------------------------
    estimate                 Optional boolean. If True, the number of credits to run the operation will be returned.
    -----------------------  -------------------------------------------------------------------------------------------
    future                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =======================  ===========================================================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else Python dictionary with the following keys:

        "similar_result_layer" : layer (FeatureCollection)

        "process_info" : list of message

    .. code-block:: python

        #USAGE EXAMPLE: To find top 4 most locations from the candidates layer that are similar to the target location.
        top_4_most_similar_locations = find_similar_locations(target_lyr, candidates_lyr,
                                                    analysis_fields=['THH17','THH35','THH02','THH05','POPDENS14','FAMGRW10_14','UNEMPRT_CY'],
                                                    output_name = "top 4 similar locations",
                                                    number_of_results=4)
    """
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.find_similar_locations(
        input_layer,
        search_layer,
        analysis_fields,
        input_query,
        number_of_results,
        output_name,
        context,
        estimate=estimate, future=future)

def find_centroids(input_layer,
                   point_location=False,
                   output_name=None,
                   context=None,
                   gis=None,
                   estimate=False,
                   future=False):
    """
    .. image:: _static/images/find_centroids/find_centroids.png

    The ``find_centroids`` method that finds and generates points from the representative center (centroid) of
    each input multipoint, line, or area feature. Finding the centroid of a feature is very common for many analytical
    workflows where the resulting points can then be used in other analytic workflows.

    For example, polygon features that contain demographic data can be converted to centroids that can be used in network analysis.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required feature layer. The multipoint, line, or polygon features that will be used to generate centroid point features. See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    point_location    Optional boolean. A Boolean value that determines the output location of the points.

                        + True - Output points will be the nearest point to the actual centroid, but located inside or contained by the bounds of the input feature.
                        + False - Output point locations will be determined by the calculated geometric center of each input feature. This is the default.
    ----------------  ---------------------------------------------------------------
    output_name       Optional string. If provided, the method will create a feature service of the results. You define the name of the service. If ``output_name`` is not supplied, the method will return a feature collection.
    ----------------  ---------------------------------------------------------------
    context           Optional string. Context contains additional settings that affect method execution. For ``find_centroids``, there are two settings.

                      #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those features in the ``input_layer`` that intersect the bounding box will be buffered.
                      #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    ----------------  ---------------------------------------------------------------
    estimate          Optional boolean. If True, the number of credits to run the operation will be returned.
    ----------------  ---------------------------------------------------------------
    future            Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    ================  ===============================================================

    :returns: result_layer : feature layer Item if ``output_name`` is specified, else Feature Collection.

    .. code-block:: python

        # USAGE EXAMPLE: To find centroids of madison fields nearest to the actual centroids.

        centroid = find_centroids(madison_fields,
                                  point_location=True,
                                  output_name='find centroids')
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    if gis._portal.is_arcgisonline == False:
        raise Exception("find_centroids is only available on ArcGIS Online.")

    return gis._tools.featureanalysis.find_centroids(input_layer,
                                                     point_location,
                                                     output_name,
                                                     context,
                                                     estimate=estimate,
                                                     future=future)



def choose_best_facilities(goal='Allocate',
                           demand_locations_layer=None,
                           demand=1,
                           demand_field=None,
                           max_travel_range=2147483647,
                           max_travel_range_field=None,
                           max_travel_range_units='Minutes',
                           travel_mode='Driving Time',
                           time_of_day=None,
                           time_zone_for_time_of_day='GeoLocal',
                           travel_direction='FacilityToDemand',
                           required_facilities_layer=None,
                           required_facilities_capacity=2147483647,
                           required_facilities_capacity_field=None,
                           candidate_facilities_layer=None,
                           candidate_count=1,
                           candidate_facilities_capacity=2147483647,
                           candidate_facilities_capacity_field=None,
                           percent_demand_coverage=100,
                           output_name=None,
                           context=None,
                           gis=None,
                           estimate=False,
                           point_barrier_layer=None,
                           line_barrier_layer=None,
                           polygon_barrier_layer=None, future=False):
    """
    .. image:: _static/images/choose_best_facilities/choose_best_facilities.png

    The ``choose_best_facilities`` method finds the set of facilities that will best serve demand from surrounding areas.

    Facilities might be public institutions that offer a service, such as fire stations, schools, or libraries,
    or they might be commercial ones such as drug stores or distribution centers for a parcel delivery service.
    Demand represents the need for a service that the facilities can meet. Demand is associated with point locations,
    with each location representing a given amount of demand.

    =====================================    =========================================================
    **Argument**                             **Description**
    -------------------------------------    ---------------------------------------------------------
    goal                                     Required string. Specify the goal that must be satisfied when allocating
                                             demand locations to facilities.

                                             Choice list:['Allocate', 'MinimizeImpedance', 'MaximizeCoverage', 'MaximizeCapacitatedCoverage', 'PercentCoverage']

                                             Default value is 'Allocate'.
    -------------------------------------    ---------------------------------------------------------
    demand_locations_layer                   Required point feature layer. A point layer specifying the locations that have demand for facilities. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    demand                                   Optional float. The amount of demand available at every demand locations.

                                             The default value is 1.0.
    -------------------------------------    ---------------------------------------------------------
    demand_field                             Optional string. A numeric field on the ``demand_locations_layer`` representing the
                                             amount of demand available at each demand location. If specified, the ``demand`` parameter is ignored.
    -------------------------------------    ---------------------------------------------------------
    max_travel_range                         Optional float. Specify the maximum travel time or distance allowed
                                             between a demand location and the facility it is allocated to.

                                             The default is unlimited (2,147,483,647.0).
    -------------------------------------    ---------------------------------------------------------
    max_travel_range_field                   Optional string. A numeric field on the ``demand_locations_layer`` specifying the maximum
                                             travel time or distance allowed between a demand location
                                             and the facility it is allocated to. If specified, the ``max_travel_range`` parameter is ignored.
    -------------------------------------    ---------------------------------------------------------
    max_travel_range_units                   Optional string. The units for the maximum travel time or distance allowed
                                             between a demand location and the facility it is allocated to.

                                             Choice list:['Seconds', 'Minutes', 'Hours', 'Days', 'Meters', 'Kilometers', 'Feet', 'Yards', 'Miles'].

                                             The default is 'Minutes'.
    -------------------------------------    ---------------------------------------------------------
    travel_mode                              Specify the mode of transportation for the analysis.

                                             Choice list: ['Driving Distance', 'Driving Time', 'Rural Driving Distance', 'Rural Driving Time', 'Trucking Distance', 'Trucking Time', 'Walking Distance', 'Walking Time']
    -------------------------------------    ---------------------------------------------------------
    time_of_day                              Optional datetime.datetime. Specify whether travel times
                                             should consider traffic conditions. To use traffic in the
                                             analysis, set travel_mode to a travel mode object whose
                                             impedance_attribute_name property is set to travel_time and
                                             assign a value to time_of_day. (A travel mode with other
                                             impedance_attribute_name values don't support traffic.)
                                             The ``time_of_day`` value represents the time at which travel
                                             begins, or departs, from the origin points. The time is
                                             specified as datetime.datetime.

                                             The service supports two kinds of traffic: ty
                                             pical and live.
                                             Typical traffic references travel speeds that are made up of
                                             historical averages for each five-minute interval spanning a week.
                                             Live traffic retrieves speeds from a traffic feed that processes
                                             phone probe records, sensors, and other data sources to record
                                             actual travel speeds and predict speeds for the near future.

                                             The `data coverage <http://www.arcgis.com/home/webmap/viewer.html?webmap=b7a893e8e1e04311bd925ea25cb8d7c7>`_ page shows the countries Esri currently provides traffic data for.

                                             Typical Traffic:

                                             To ensure the task uses typical traffic in locations where it
                                             is available, choose a time and day of the week, and then convert
                                             the day of the week to one of the following dates from 1990:

                                             * Monday - 1/1/1990
                                             * Tuesday - 1/2/1990
                                             * Wednesday - 1/3/1990
                                             * Thursday - 1/4/1990
                                             * Friday - 1/5/1990
                                             * Saturday - 1/6/1990
                                             * Sunday - 1/7/1990
                                             Set the time and date as datetime.datetime.

                                             For example, to solve for 1:03 p.m. on Thursdays, set the
                                             time and date to 1:03 p.m., 4 January 1990; and convert to
                                             datetime eg. datetime.datetime(1990, 1, 4, 1, 3).

                                             Live Traffic:

                                             To use live traffic when and where it is available,
                                             choose a time and date and convert to datetime.

                                             Esri saves live traffic data for 12 hours and references
                                             predictive data extending 12 hours into the future. If the
                                             time and date you specify for this parameter is outside the
                                             24-hour time window, or the travel time in the analysis
                                             continues past the predictive data window, the task falls
                                             back to typical traffic speeds.

                                             Examples:
                                             from datetime import datetime

                                             * "time_of_day": datetime(1990, 1, 4, 1, 3) # 13:03, 4 January 1990. Typical traffic on Thursdays at 1:03 p.m.
                                             * "time_of_day": datetime(1990, 1, 7, 17, 0) # 17:00, 7 January 1990. Typical traffic on Sundays at 5:00 p.m.
                                             * "time_of_day": datetime(2014, 10, 22, 8, 0) # 8:00, 22 October 2014. If the current time is between 8:00 p.m., 21 Oct. 2014 and 8:00 p.m., 22 Oct. 2014, live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
                                             * "time_of_day": datetime(2015, 3, 18, 10, 20) # 10:20, 18 March 2015. If the current time is between 10:20 p.m., 17 Mar. 2015 and 10:20 p.m., 18 Mar. 2015, live traffic speeds are referenced in the analysis; otherwise, typical traffic speeds are referenced.
    -------------------------------------    ---------------------------------------------------------
    time_zone_for_time_of_day                Optional string. Specify the time zone or zones of the time_of_day parameter.

                                             Choice list: ['GeoLocal', 'UTC']

                                             GeoLocal-refers to the time zone in which the origins_layer points are located.

                                             UTC-refers to Coordinated Universal Time.
    -------------------------------------    ---------------------------------------------------------
    travel_direction                         Optional string. Specify whether to measure travel times or distances
                                             from facilities to demand locations or from demand locations to facilities.

                                             Choice list:['FacilityToDemand', 'DemandToFacility']
    -------------------------------------    ---------------------------------------------------------
    required_facilities_layer                Optional point feature layer. A point layer specifying one or more locations that act as facilities
                                             by providing some kind of service. Facilities specified by this parameter
                                             are required to be part of the output solution and will be used before any
                                             facilities from the ``candidate_facilities_layer`` when allocating demand locations.
    -------------------------------------    ---------------------------------------------------------
    required_facilities_capacity             Optional float. Specify how much demand every facility in the ``required_facilities_layer`` is capable of supplying.

                                             The default value is unlimited (2,147,483,647).
    -------------------------------------    ---------------------------------------------------------
    required_facilities_capacity_field       Optional string. A field on the required_facilities_layer representing how much demand each facility in the ``required_facilities_layer`` is capable of supplying. This parameter takes precedence when ``required_facilities_capacity`` parameter is also specified.
    -------------------------------------    ---------------------------------------------------------
    candidate_facilities_layer               Optional point layer. A point layer specifying one or more locations that act as facilities by providing some kind of service. Facilities specified by this parameter are not required to be part of the output solution and will be used only after all the facilities from the ``candidate_facilities_layer`` have been used when allocating demand locations.
    -------------------------------------    ---------------------------------------------------------
    candidate_count                          Optional integer. The number of candidate facilities to choose when allocating demand locations. Note that the sum of the features in the ``required_facilities_capacity`` and the value specified for ``candidate_count`` cannot exceed 100.

                                             The default value is 1.
    -------------------------------------    ---------------------------------------------------------
    candidate_facilities_capacity            Optional float. Specify how much demand every facility in the ``candidate_facilities_layer`` is capable of supplying.

                                             The default value is unlimited (2,147,483,647.0).
    -------------------------------------    ---------------------------------------------------------
    candidate_facilities_capacity_field      Optional string. A field on the ``candidate_facilities_layer`` representing how much demand each facility in the ``candidate_facilities_layer`` is capable of supplying. This parameter takes precedence when ``candidate_facilities_capacity`` parameter is also specified.
    -------------------------------------    ---------------------------------------------------------
    percent_demand_coverage                  Optional float. Specify the percentage of the total demand that you want the chosen and required facilities to capture.

                                             The default value is 100.
    -------------------------------------    ---------------------------------------------------------
    output_name                              Optional string. If provided, the method will create a feature layer of the results. You define the name of the layer. If ``output_name`` is not supplied, the method will return a feature collection.
    -------------------------------------    ---------------------------------------------------------
    context                                  Optional string. Additional settings such as processing extent and output spatial reference. For ``choose_best_facilities``, there are two settings.

                                             #. Extent (``extent``)-a bounding box that defines the analysis area. Only those features in the input layer that intersect the bounding box will be enriched.
                                             #. Output Spatial Reference (``outSR``)-the output features will be projected into the output spatial reference.
    -------------------------------------    ---------------------------------------------------------
    gis                                      Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------------------    ---------------------------------------------------------
    estimate                                 Optional boolean. Is true, the number of credits needed to run the operation will be returned as a float.
    -------------------------------------    ---------------------------------------------------------
    point_barrier_layer                      Optional layer. Specify one or more point features that act as temporary restrictions (in other words, barriers) when traveling on the underlying streets.

                                             A point barrier can model a fallen tree, an accident, a downed electrical line, or anything that completely blocks traffic at a specific position along
                                             the street. Travel is permitted on the street but not through the barrier. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    line_barrier_layer                       Optional layer. Specify one or more line features that prohibit travel anywhere the lines intersect the streets.

                                             A line barrier prohibits travel anywhere the barrier intersects the streets. For example, a parade or protest that blocks traffic across several street
                                             segments can be modeled with a line barrier. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    polygon_barrier_layer                    Optional string. Specify one or more polygon features that completely restrict travel on the streets intersected by the polygons.

                                             One use of this type of barrier is to model floods covering areas of the street network and making road travel there impossible. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ------------------------------------------------------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================    =========================================================


    :returns: When an output_name is specified, a FeatureLayerCollection Item with 3 layers is returned (see dictionary below for details), else a dict with the following keys:

       "allocated_demand_locations_layer" : layer (FeatureCollection)

       "allocation_lines_layer"  : layer (FeatureCollection)

       "assigned_facilities_layer"   : layer (FeatureCollection)

    .. code-block:: python

        USAGE EXAMPLE: To minimize overall distance travelled for travelling from esri offices to glider airports.

        best_facility = choose_best_facilities(goal="MinimizeImpedance",
                                    demand_locations_layer=esri_offices,
                                    travel_mode='Driving Distance',
                                    travel_direction="DemandToFacility",
                                    required_facilities_layer=gliderport_lyr,
                                    candidate_facilities_layer=balloonport_lyr,
                                    candidate_count=1,
                                    output_name="choose best facilities")
    """
    gis = _arcgis.env.active_gis if gis is None else gis

    if isinstance(travel_mode, str):
        route_service = network.RouteLayer(gis.properties.helperServices.route.url, gis=gis)
        travelmodes = route_service.retrieve_travel_modes()
        for tm in travelmodes['supportedTravelModes']:
            if tm['name'] == travel_mode:
              tm = [i for i in route_service.retrieve_travel_modes()['supportedTravelModes'] if i['name'] == travel_mode][0]
              travel_mode = tm

    return gis._tools.featureanalysis.choose_best_facilities(
        goal,
        demand_locations_layer,
        demand,
        demand_field,
        max_travel_range,
        max_travel_range_field,
        max_travel_range_units,
        travel_mode,
        time_of_day,
        time_zone_for_time_of_day,
        travel_direction,
        required_facilities_layer,
        required_facilities_capacity,
        required_facilities_capacity_field,
        candidate_facilities_layer,
        candidate_count,
        candidate_facilities_capacity,
        candidate_facilities_capacity_field,
        percent_demand_coverage,
        output_name,
        context,
        estimate=estimate,
        point_barrier_layer=point_barrier_layer,
        line_barrier_layer=line_barrier_layer,
        polygon_barrier_layer=polygon_barrier_layer,
        future=future)

def create_viewshed(
        input_layer,
        dem_resolution="Finest",
        maximum_distance=None,
        max_distance_units="Meters",
        observer_height=None,
        observer_height_units="Meters",
        target_height=None,
        target_height_units="Meters",
        generalize=True,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/create_viewshed/create_viewshed.png

    The create_viewshed method identifies visible areas based on the observer locations you provide.
    The results are areas where the observers can see the observed objects (and the observed objects can see the observers).

    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required point feature layer. The features to use as the observer locations. See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    dem_resolution               Optional string. The approximate spatial resolution (cell size) of the source elevation data used for the calculation.

                                 The resolution values are an approximation of the spatial resolution of the digital elevation model.
                                 While many elevation sources are distributed in units of arc seconds, the keyword is an approximation
                                 of those resolutions in meters for easier understanding.

                                 Choice list: ['FINEST', '10m', '24m', '30m', '90m']

                                 The default is the finest resolution available.
    -------------------------    ---------------------------------------------------------
    maximum_distance             Optional float. This is a cutoff distance where the computation of visible areas stops. Beyond this distance, it is unknown whether the analysis points and the other objects can see each other.

                                 It is useful for modeling current weather conditions or a given time of day, such as dusk. Large values increase computation time.

                                 Unless specified, a default maximum distance will be computed based on the resolution and extent of the source DEM. The allowed maximum value is 50 kilometers.
                                 Use max_distance_units to set the units for maximum_distance.
    -------------------------    ---------------------------------------------------------
    max_distance_units           Optional string. The units for the maximum_distance parameter.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'Yards']

                                 The default is 'Meters'.
    -------------------------    ---------------------------------------------------------
    observer_height              Optional float. This is the height above the ground of the observer locations.

                                 The default is 1.75 meters, which is approximately the average height of a person. If you are looking from an elevated location, such as an observation tower or a tall building, use that height instead.

                                 Use observer_height_units to set the units for observer_height.

    -------------------------    ---------------------------------------------------------
    observer_height_units        Optional string. The units for the observer_height parameter.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'Yards']

                                 The default is 'Meters'.
    -------------------------    ---------------------------------------------------------
    target_height                Optional float. This is the height of structures or people on the ground used to
                                 establish visibility. The result viewshed are those areas where an input point can see these other objects.
                                 The converse is also true; the other objects can see an input point.

                                 * If your input points represent wind turbines and you want to determine where people standing on the
                                   ground can see the turbines, enter the average height of a person (approximately 6 feet).
                                   The result is those areas where a person standing on the ground can see the wind turbines.
                                 * If your input points represent fire lookout towers and you want to determine which lookout
                                   towers can see a smoke plume 20 feet high or higher, enter 20 feet for the height. The result
                                   is those areas where a fire lookout tower can see a smoke plume at least 20 feet high.
                                 * If your input points represent scenic overlooks along roads and trails and you want to determine
                                   where wind turbines 400 feet high or higher can be seen, enter 400 feet for the height. The result
                                   is those areas where a person standing at a scenic overlook can see a wind turbine at least 400 feet high.
                                 * If your input points represent scenic overlooks and you want to determine how much area on the ground
                                   people standing at the overlook can see, enter zero. The result is those areas that can be seen from the scenic overlook.

                                 Use target_height_units to set the units for target_height.

    -------------------------    ---------------------------------------------------------
    target_height_units          Optional string. The units for the target_height parameter.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'Yards']

                                 The default is 'Meters'.
    -------------------------    ---------------------------------------------------------
    generalize                   Optional boolean. Determines whether or not the viewshed polygons are to be generalized.

                                 The viewshed calculation is based on a raster elevation model that creates a result with stair-stepped edges.
                                 To create a more pleasing appearance and improve performance, the default behavior is to generalize the polygons.
                                 The generalization process smooths the boundary of the visible areas and may remove some single-cell visible areas.

                                 The default value is True.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Output feature service name. If not provided, a feature collection is returned.
    -------------------------    ---------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. For ``create_viewshed``, there are two settings.

                                 #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the ``input_layer``
                                    that intersect the bounding box will be analyzed.

                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional boolean. If True, the estimated number of credits required to run the operation will be returned.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    =========================================================

    :returns result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        USAGE EXAMPLE: To create viewshed around esri headquarter office.

        viewshed3 = create_viewshed(hq_lyr,
                            maximum_distance=9,
                            max_distance_units='Miles',
                            target_height=6,
                            target_height_units='Feet',
                            output_name="create Viewshed")

    """
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.create_viewshed(
        input_layer,
        dem_resolution,
        maximum_distance,
        max_distance_units,
        observer_height,
        observer_height_units,
        target_height,
        target_height_units,
        generalize,
        output_name,
        context,
        estimate=estimate, future=future)


def create_watersheds(
        input_layer,
        search_distance=None,
        search_units="Meters",
        source_database="FINEST",
        generalize=True,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/create_watersheds/create_watersheds.png

    The ``create_watersheds`` method determines the watershed, or upstream contributing area, for each point
    in your analysis layer. For example, suppose you have point features representing locations
    of waterborne contamination, and you want to find the likely sources of the contamination.
    Since the source of the contamination must be somewhere within the watershed upstream of the
    point, you would use this tool to define the watersheds containing the sources of the contaminant.


    =========================    =========================================================
    **Parameter**                **Description**
    -------------------------    ---------------------------------------------------------
    input_layer                  Required point feature layer. The point features used for calculating watersheds.
                                 These are referred to as pour points, because it is the location at which water pours out of the watershed.
                                 See :ref:`Feature Input<FeatureInput>`.
    -------------------------    ---------------------------------------------------------
    search_distance              Optional float. The maximum distance to move the location of an input point.
                                 Use search_units to set the units for search_distance.

                                 If your input points are located away from a drainage line, the resulting watersheds
                                 are likely to be very small and not of much use in determining the upstream source of
                                 contamination. In most cases, you want your input points to snap to the nearest drainage
                                 line in order to find the watersheds that flows to a point located on the drainage line.
                                 To find the closest drainage line, specify a search distance. If you do not specify a
                                 search distance, the tool will compute and use a conservative search distance.

                                 To use the exact location of your input point, specify a search distance of zero.

                                 For analysis purposes, drainage lines have been precomputed by Esri using standard
                                 hydrologic models. If there is no drainage line within the search distance, the location
                                 containing the highest flow accumulation within the search distance is used.
    -------------------------    ---------------------------------------------------------
    search_units                 Optional string. The linear units specified for the search distance.

                                 Choice list: ['Meters', 'Kilometers', 'Feet', 'Miles', 'Yards']
    -------------------------    ---------------------------------------------------------
    source_database              Optional string. Keyword indicating the data source resolution that will be used in the analysis.

                                 Choice list: ['Finest', '30m', '90m']

                                 * Finest (Default): Finest resolution available at each location from all possible data sources.
                                 * 30m: The hydrologic source was built from 1 arc second - approximately 30 meter resolution, elevation data.
                                 * 90m: The hydrologic source was built from 3 arc second - approximately 90 meter resolution, elevation data.
    -------------------------    ---------------------------------------------------------
    generalize                   Optional boolean. Determines if the output watersheds will be smoothed into simpler shapes or conform
                                 to the cell edges of the original DEM.

                                 * True: The polygons will be smoothed into simpler shapes. This is the default.
                                 * False: The edge of the polygons will conform to the edges of the original DEM.

                                 The default value is True.
    -------------------------    ---------------------------------------------------------
    output_name                  Optional string. Output feature service name. If not provided, a feature collection is returned.
    -------------------------    ---------------------------------------------------------
    context                      Optional dict. Context contains additional settings that affect task execution. For ``create_watersheds``, there are two settings.

                                 #. Extent (``extent``)-a bounding box that defines the analysis area. Only those points in the ``input_layer``
                                    that intersect the bounding box will be analyzed.

                                 #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------    ---------------------------------------------------------
    gis                          Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    -------------------------    ---------------------------------------------------------
    estimate                     Optional boolean. If True, the estimated number of credits required to run the operation will be returned.
    -------------------------    ---------------------------------------------------------
    future                       Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =========================    =========================================================

    :returns result_layer : feature layer Item if output_name is specified, else Feature Collection.

    .. code-block:: python

        USAGE EXAMPLE: To create watersheds for Chennai lakes.

        lakes_watershed = create_watersheds(lakes_lyr,
                                            search_distance=3,
                                            search_units='Kilometers',
                                            source_database='90m',
                                            output_name='create watersheds')

    """
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.create_watersheds(
        input_layer,
        search_distance,
        search_units,
        source_database,
        generalize,
        output_name,
        context,
        estimate=estimate, future=future)


def trace_downstream(
        input_layer,
        split_distance=None,
        split_units="Kilometers",
        max_distance=None,
        max_distance_units="Kilometers",
        bounding_polygon_layer=None,
        source_database=None,
        generalize=True,
        output_name=None,
        context=None,
        gis=None,
        estimate=False,
        future=False):
    """
    .. image:: _static/images/trace_downstream/trace_downstream.png

    The ``trace_downstream`` method determines the trace, or flow path, in a downstream direction from the points in your analysis layer.

    For example, suppose you have point features representing sources of contamination and you want to determine where in your study
    area the contamination will flow. You can use ``trace_downstream`` to identify the path the contamination will take. This trace
    can also be divided into individual line segments by specifying a distance value and units. The line being returned can be the
    total length of the flow path, a specified maximum trace length, or clipped to area features such as your study area. In many
    cases, if the total length of the trace path is returned, it will be from the source all the way to the ocean.

    =====================================    =========================================================
    **Argument**                             **Description**
    -------------------------------------    ---------------------------------------------------------
    input_layer                              Required feature layer. The point features used for the starting location of a downstream trace.
                                             See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    split_distance                           Optional float. The trace line will be split into multiple lines where each line is of the specified length.
                                             The resulting trace will have multiple line segments, each with fields FromDistance and ToDistance.
    -------------------------------------    ---------------------------------------------------------
    split_units                              Optional string. The units used to specify split distance.

                                             Choice list: ['Meters', 'Kilometers', 'Feet' 'Yards', 'Miles'].

                                             The default is 'Kilometers'.
    -------------------------------------    ---------------------------------------------------------
    max_distance                             Optional float. Determines the total length of the line that will be returned. If you provide a
                                             ``bounding_polygon_layer`` to clip the trace, the result will be clipped to the features in ``bounding_polygon_layer``,
                                             regardless of the distance you enter here.
    -------------------------------------    ---------------------------------------------------------
    max_distance_units                       Optional string. The units used to specify maximum distance.

                                             Choice list: ['Meters', 'Kilometers', 'Feet' 'Yards', 'Miles'].

                                             The default is 'Kilometers'.
    -------------------------------------    ---------------------------------------------------------
    bounding_polygon_layer                   Optional feature layer. A polygon layer specifying the area(s) where you want the trace
                                             downstreams to be calculated in. For example, if you only want to calculate the trace downstream
                                             with in a county polygon, provide a layer containing the county polygon and the resulting trace
                                             lines will be clipped to the county boundary. See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------    ---------------------------------------------------------
    source_database                          Optional string. Keyword indicating the data source resolution that will be used in the analysis.

                                             Choice list: ['Finest', '30m', '90m'].

                                                * Finest: Finest resolution available at each location from all possible data sources.

                                                * 30m: The hydrologic source was built from 1 arc second - approximately 30 meter resolution, elevation data.

                                                * 90m: The hydrologic source was built from 3 arc second - approximately 90 meter resolution, elevation data.

                                             The default is 'Finest'.
    -------------------------------------    ---------------------------------------------------------
    generalize                               Optional boolean. Determines if the output trace downstream lines will be smoothed
                                             into simpler lines or conform to the cell edges of the original DEM.
    -------------------------------------    ---------------------------------------------------------
    output_name                              Optional string. If provided, the task will create a feature service of the results.
                                             You define the name of the service. If ``output_name`` is not supplied, the task will return a feature collection.
    -------------------------------------    ---------------------------------------------------------
    context                                  Optional string. Context contains additional settings that affect task execution. For ``trace_downstream``, there are two settings.

                                             #.  Extent (``extent``) - a bounding box that defines the analysis area. Only those points
                                                in the ``input_layer`` that intersect the bounding box will have a downstream trace generated.
                                             #. Output Spatial Reference (``outSR``) - the output features will be projected into the output spatial reference.
    -------------------------------------    ---------------------------------------------------------
    estimate                                 Optional boolean. If True, the number of credits to run the operation will be returned.
    -------------------------------------    ---------------------------------------------------------
    future                                   Optional boolean. If True, the result will be a GPJob object and results will be returned asynchronously.
    =====================================    =========================================================

    :returns: feature layer collection if ``output_name`` is set, else feature collection.

    .. code-block:: python

        # USAGE EXAMPLE: To identify the path the water contamination  will take.
        path = trace_downstream(input_layer=water_source_lyr,
                                split_distance=2,
                                split_units='Miles',
                                max_distance=2,
                                max_distance_units='Miles',
                                source_database='Finest',
                                generalize=True,
                                output_name='trace downstream')
    """
    gis = _arcgis.env.active_gis if gis is None else gis

    return gis._tools.featureanalysis.trace_downstream(
        input_layer,
        split_distance,
        split_units,
        max_distance,
        max_distance_units,
        bounding_polygon_layer,
        source_database,
        generalize,
        output_name,
        context,
        estimate=estimate, future=future)



"""

These tools are used for the day-to-day management of geographic and tabular data.

copy_to_data_store copies data to your ArcGIS Data Store and creates a layer in your web GIS.
"""
import json as _json
import logging as _logging
import datetime as _datetime
import arcgis as _arcgis
from arcgis.features import FeatureSet as _FeatureSet
from arcgis.geoprocessing._support import _execute_gp_tool
from ._util import _id_generator, _feature_input, _set_context, _create_output_service, GAJob

_log = _logging.getLogger(__name__)

_use_async = True

def run_python_script(code, layers=None, gis=None, context=None, future=False):
    """

    The ``run_python_script`` method executes a Python script on your ArcGIS
    GeoAnalytics Server site. In the script, you can create an analysis
    pipeline by chaining together multiple GeoAnalytics Tools without
    writing intermediate results to a data store. You can also use other
    Python functionality in the script that can be distributed across your
    GeoAnalytics Server.

    For example, suppose that each week you receive a new dataset of
    vehicle locations containing billions of point features. Each time you
    receive a new dataset, you must perform the same workflow involving
    multiple GeoAnalytics Tools to create an information product that you
    share within your organization. This workflow creates several large
    intermediate layers that take up lots of space in your data store. By
    scripting this workflow in Python and executing the code in the Run
    Python Script task, you can avoid creating these unnecessary
    intermediate layers, while simplifying the steps to create the
    information product.

    When you use ``run_python_script``, the Python code is executed on your
    GeoAnalytics Server. The script runs with the Python 3.6 environment
    that is installed with GeoAnalytics Server, and all console output is
    returned as job messages. Some Python modules can be used in your
    script to execute code across multiple cores of one or more machines
    in your GeoAnalytics Server using Spark 2.2.0(the compute platform that
    distributes analysis for GeoAnalytics Tools).

    A geoanalytics module is available and allows you to run GeoAnalytics
    Tools in the script. This package is imported automatically when you
    use Run Python Script.

    To interact directly with Spark in the Run Python Script task, use the
    pyspark module, which is imported automatically when you run the task.
    The pyspark module is the Python API for Spark and provides a
    collection of distributed analysis tools for data management,
    clustering, regression, and more that can be called in Run Python
    Script and run across your GeoAnalytics Server.

    When using the geoanalytics and pyspark packages, most functions return
    analysis results in memory as Spark DataFrames. Spark data frames can be
    written to a data store or used in the script. This allows for the
    chaining together of multiple geoanalytics and pyspark tools, while only
    writing out the final result to a data store, eliminating the need to
    create any intermediate result layers.

    For advanced users, an instance of SparkContext is instantiated
    automatically as sc and can be used in the script to interact with Spark.
    This allows for the execution of custom distributed analysis across your
    GeoAnalytics Server.

    It is recommended that you use an integrated development environment
    (IDE) to write your Python script, and copy the script text into the Run
    Python Script tool. This makes it easier to identify syntax errors and
    typos prior to running your script. It is also recommended that you run
    your script using a small subset of the input data first to verify that
    there are no logic errors or exceptions. You can use the Describe
    Dataset task to create a sample layer for this purpose.

    ================  ===============================================================
    code              Required string. The Python script that will run on your GeoAnalytics Server. This must be the full script as a string.

                      The layers provided in inputLayers can be accessed in the script using the layers object. To learn more, 
                      see `Reading and writing layers in pyspark <https://developers.arcgis.com/rest/services-reference/using-webgis-layers-in-pyspark.htm>`_.
  
                      GeoAnalytics Tools can be accessed with the geoanalytics object, which is instantiated in the script environment automatically. 
                      To learn more, see `Using GeoAnalytics Tools in Run Python Script <https://developers.arcgis.com/rest/services-reference/using-geoanalytics-tools-in-pyspark.htm>`_.
  
                      For a collection of example scripts, see `Examples: Scripting custom analysis with the Run Python Script task <https://developers.arcgis.com/rest/services-reference/run-python-script-examples.htm>`_.
    ----------------  ---------------------------------------------------------------
    layers            Optional list. A list of Feature layers to operate on. See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ----------------  ---------------------------------------------------------------
    context           Optional dict. This parameter is not used by the ``run_python_script`` tool.

                      To control the output data store, use the "dataStore" option when writing DataFrames.
  
                      To set the processing or output spatial reference, use the project tool in the geoanalytics package.
  
                      To filter a layer when converting it to a DataFrame, use the "where" or "fields" option when loading the layer's URL.
  
                      To limit the extent of a layer when converting it to a DataFrame, use the "extent" option when loading the layer's URL.
    ----------------  ---------------------------------------------------------------
    future            Optional boolean. If 'True', a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.
    ================  ===============================================================

    :returns: list of dictionary of messages from the code provided.

    .. code-block:: python

            # Usage Example: Execute calculate_density tool using run_python_script.

            code = '''
            import time 
            r = geoanalytics.calculate_density(input_layer=layers[0],
                            fields='Damage',
                            weight='Uniform',
                            bin_type='Square',
                            bin_size=1,
                            bin_size_unit='Meters',
                            radius=2,
                            radius_unit='Yards')        
                            
            r.write.format('webgis').save('RunPythonScriptTest_{0}'.format(time.time()))              
            '''
            run_python_script(code=code, layers=[lyr])
    """
    if layers is None:
        layers = []
    import inspect
    params = {'f': 'json'}

    if inspect.isfunction(code):
        params['code'] = inspect.getsource(code) + '\n' + code.__name__ + '()'
    elif isinstance(code, str):
        params['code'] = code
    else:
        raise ValueError("code must be a string or Python Function.")

    if isinstance(layers, (tuple, list)):
        params['layers'] = layers
    else:
        raise ValueError("layers must be a list or tuple")

    tool_name = "RunPythonScript"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "layers": (_FeatureSet, "inputLayers"),
        "code" : (str, "pythonScript"),
        "context": (str, "context"),
    }

    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, [], _use_async, url, True, return_messages=False, future=future)
            return GAJob(gpjob=gpjob, return_service=None)
        res, msg = _execute_gp_tool(gis, tool_name, params, param_db, [], _use_async, url, True, return_messages=True, future=future)
        msg = msg['messages']
        return msg
    except:
        raise
    return 

def dissolve_boundaries(input_layer,
                        dissolve_fields=None,
                        summary_fields=None,
                        multipart=False,
                        output_name=None,
                        gis=None,
                        context=None,
                        future=False):
    """
    .. image:: _static/images/dissolve_boundaries/dissolve_boundaries.png 

    The ``dissolve_boundaries`` task finds polygons that intersect or have the same field values 
    and merges them together to form a single polygon.

    Example:

        A city council wants to control liquor sales by refusing new licenses to stores within 
        1,000 feet of schools, libraries, and parks. After creating a 1,000-foot buffer around 
        the schools, libraries, and parks, the buffered layers can be joined together and the 
        boundaries can be dissolved to create a single layer of restricted areas.

    .. note::
        Only available at ArcGIS Enterprise 10.7 and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required layer. The layer containing polygon features that will be dissolved. See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    dissolve_fields   Optional string. A comma seperated list of strings for each
                      field that you want to dissolve on.One or more fields in the ``input_layer`` that determine how polygons are merged based on field value.

                      If you don't specify fields, polygons that intersect will be dissolved into one polygon by default.

                      If you do specify fields, polygons that share the same value for each of the specified fields will be dissolved into one polygon.
    ----------------  ---------------------------------------------------------------
    summary_fields    Optional list of dicts. A list of field names and statistical summary types you want to calculate. 
                      Note that the count is always returned. By default, all statistics are returned.

                      Syntax: [{"statisticType" : "<stat>", "onStatisticField" : "<field name>"}]

                      fieldName is the name of the fields in the input point layer.

                      statisticType is one of the following for numeric fields:

                        * ``Count`` - Totals the number of values of all the points in each polygon.
                        * ``Sum`` - Adds the total value of all the points in each polygon.
                        * ``Mean`` - Calculates the average of all the points in each polygon.
                        * ``Min`` - Finds the smallest value of all the points in each polygon.
                        * ``Max`` - Finds the largest value of all the points in each polygon.
                        * ``Range`` - Finds the difference between the Min and Max values.
                        * ``Stddev`` - Finds the standard deviation of all the points in each polygon.
                        * ``Var`` - Finds the variance of all the points in each polygon.
                        statisticType is one of the following for string fields:

                        * ``Count`` - Totals the number of strings for all the points in each polygon.
                        * ``Any`` - Returns a sample string of a point in each polygon.
                  
                      Example: summary_fields = [{"statisticType" : "Sum", "onStatisticField" : "quadrat_area_km2"}, {"statisticType" : "Mean", "onStatisticField" : "soil_depth_cm"}, {"statisticType" : "Any", "onStatisticField" : "quadrat_desc"}]
    ----------------  ---------------------------------------------------------------
    multipart         Optional boolean. If 'True', the output service can contain
                      multipart features. If 'False', the output service
                      will only contain single-part features, and individual features
                      will be created for each part.

                      The default value is 'False'.
    ----------------  ---------------------------------------------------------------
    output_name       optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ----------------  ---------------------------------------------------------------
    context           Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are five settings:

                      #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                      #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                      #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                      #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
                      #. Default aggregation styles (``defaultAggregationStyles``) - If set to true, results will have square, hexagon, and triangle aggregation styles enabled on results map services.
    ----------------  ---------------------------------------------------------------
    future            optional boolean. If True, a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.
    ================  ===============================================================

    :returns: result_layer : Output Features as feature layer collection item.

    .. code-block:: python

            # Usage Example: This example dissolves boundaries of soil areas in Nebraska if they have 
            # the same solubility. For dissolved features, it calculates the sum of the quadrat area, 
            # the mean soil depth, and an example of the quadrat description.
            
            arcgis.env.out_spatial_reference = 3310
            arcgis.env.output_datastore= "relational"
            arcgis.env.defaultAggregations= True

            summary_fields = [{"statisticType" : "Sum", "onStatisticField" : "quadrat_area_km2"},
                            {"statisticType" : "Mean", "onStatisticField" : "soil_depth_cm"},
                            {"statisticType" : "Any", "onStatisticField" : "quadrat_desc"}]

            dissolve_result = dissolve_boundaries(input_layer=study_area_lyr,
                                                  dissolve_fields="soil_suitability",
                                                  summary_fields=summary_fields,
                                                  multipart=True,
                                                  output_name="Soil_Suitability_dissolved")
    """
    kwargs = locals()
    tool_name = "DissolveBoundaries"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Dissolve_Bounds_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "multipart": (bool, "multipart"),
        "summary_fields": (list, "summaryFields"),
        "dissolve_fields" : (list, "dissolveFields"),
        "output_name": (str, "OutputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }

    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]

    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise
    return

def merge_layers(input_layer, 
                 merge_layer,
                 merge_attributes=None, 
                 output_name=None,
                 gis=None, 
                 context=None,
                 future=False):
    """

    .. image:: _static/images/merge_layers/merge_layers.png 
    
    The ``merge_layers`` task combines two feature layers to create a single output layer. The tool
    requires that both layers have the same geometry type (tabular, point, line, or polygon). If
    time is enabled on one layer, the other must also be time enabled and have the same time type
    (instant or interval). The result will always contain all fields from the input layer. All
    fields from the merge layer will be included by default, or you can specify custom merge rules
    to define the resulting schema. For example:

        * I have three layers for England, Wales, and Scotland and I want a single layer of Great
          Britain. I can use Merge Layers to combine the areas and maintain all fields from each area.
        * I have two layers containing parcel information for contiguous townships. I want to join them
          together into a single layer, keeping only the fields that have the same name and type in the
          two layers.

    .. note::
        Only available at **ArcGIS Enterprise 10.7** and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required layer. The table, point, line, or polygon features to 
                      merge with the ``merge_layer`` parameter. All fields in ``input_layer`` 
                      will be included in the result layer.  See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    merge_layer       Required layer. The point, line, or polygon features to
                      merge with the ``input_layer``. The ``merge_layer`` must contain the
                      same geometry type (tabular, point, line, or polygon) and the
                      same time type (none, instant, or interval) as the ``input_layer``.
                      All fields in the ``merge_layer`` will be included in the result
                      layer by default or you can define ``merge_attributes`` to
                      customize the resulting schema.  See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    merge_attributes  Optional list of dicts. Defines how the fields in ``merge_layer`` will be
                      modified. By default, all fields from both inputs will be
                      included in the output layer.

                      If a field exists in one layer but not the other, the output
                      layer will still contain the field. The output field will
                      contain null values for the input features that did not have the
                      field. For example, if the ``input_layer`` contains a field named
                      TYPE but the ``merge_layer`` does not contain TYPE, the output will
                      contain TYPE, but its values will be null for all the features
                      copied from the ``merge_layer``.

                      You can control how fields in the ``merge_layer`` are written to the
                      output layer using the following merge types that operate on a
                      specified ``merge_layer`` field:

                            * ``Remove`` - The field in the ``merge_layer`` will be removed from the output layer.
                            * ``Rename`` - The field in the ``merge_layer`` will be renamed in the output layer. 
                              You cannot rename a field in the ``merge_layer`` to a field in the ``input_layer``. 
                              If you want to make field names equivalent, use Match.
                            * ``Match`` - A field in the merge_layer is made equivalent to a field in the ``input_layer`` 
                              specified by ``merge_layer``. For example, the input_layer has a field named CODE 
                              and the merge_layer has a field named STATUS. You can match STATUS to CODE, and 
                              the output will contain the CODE field with values of the STATUS field used for 
                              features copied from the merge_layer. Type casting is supported (for example, 
                              double to integer, integer to string) except for string to numeric.
                    
                      Example: [{"mergeLayerField": "Mean_Sales","mergeType": "Match","mergeValue": "Average_Sales"},{"mergeLayerField": "Bonus","mergeType": "Remove",},{"mergeLayerField": "Field4","mergeType": "Rename","mergeValue": "Errors"}]
    ----------------  ---------------------------------------------------------------
    output_name       Optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    gis               Optional GIS. The GIS object where the analysis will take place.
    ----------------  ---------------------------------------------------------------
    context           Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are five settings:

                      #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                      #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                      #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                      #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
                      #. Default aggregation styles (``defaultAggregationStyles``) - If set to 'True', results will have square, hexagon, and triangle aggregation styles enabled on results map services.
    ----------------  ---------------------------------------------------------------    
    future            Optional boolean. If 'True', a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.

                      The default value is 'False'.
    ================  ===============================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To merge census blocks from two states into one output layer.

            merge_result = merge_layers(input_layer=il_block,
                                       merge_layer=wi_block,
                                       merge_attributes=[{"mergeLayerField" : "State_Code", "mergeType" : "Match", "mergeValue" : "statecode"}],
                                       output_name="IL_WI_Census_Blocks")
    """
    kwargs = locals()
    tool_name = "MergeLayers"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():

        if value is not None:
            params[key] = value
        elif key == 'merge_attributes' and value is None:
            params[key] = []
    if output_name is None:
        output_service_name = 'Merge_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Merge Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "merge_layer": (_FeatureSet, "mergeLayer"),
        "merge_attributes" : (list, "mergingAttributes"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return


def clip_layer(input_layer, clip_layer, output_name=None, gis=None, context=None, future=False):
    """
    .. image:: _static/images/clip_layer/clip_layer.png 

    ``clip_layer`` features from one layer to the extent of a boundary layer. Use this tool to cut out a piece
    of one feature class using one or more of the features in another feature class as a cookie
    cutter. This is particularly useful for creating a new feature layers - also referred to as study
    area or area of interest (AOI) - that contains a geographic subset of the features in another,
    larger feature class.

    Only available at **ArcGIS Enterprise 10.7** and later.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       Required feature layer. The point, line, or polygon features 
                      that will be clipped to the areas of ``clip_layer`` features. 
                      See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    clip_layer        Required feature layer. The polygon features that define the 
                      areas to which ``input_layer`` features will be clipped. 
                      See :ref:`Feature Input<FeatureInput>`.
    ----------------  ---------------------------------------------------------------
    output_name       Optional string. The task will create a feature service of 
                      the results. You define the name of the service.
    ----------------  ---------------------------------------------------------------
    context           Optional strin. The context parameter contains additional 
                      settings that affect task execution. For this task, there are four settings:

                      #. Extent (``extent``) - A bounding box that defines the analysis area. 
                         Only those features that intersect the bounding box will be analyzed.
                      #. Processing spatial reference (``processSR``) - The features will be 
                         projected into this coordinate system for analysis.
                      #. Output spatial reference (``outSR``) - The features will be projected 
                         into this coordinate system after the analysis to be saved. 
                         The output spatial reference for the spatiotemporal big data store is always WGS84.
                      #. Data store (``dataStore``) - Results will be saved to the specified data store. 
                         The default is the spatiotemporal big data store.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS. The GIS object where the analysis will take place.
    ----------------  ---------------------------------------------------------------
    future            Optional boolean. If True, a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.
    ================  ===============================================================

    :returns: feature layer collection

    .. code-block:: python

            # Usage Example: To clip the buffered area in the shape of Capitol Hill boundary.

            clipped = clip_layer(input_layer=buffer, 
                                 clip_layer=boundary,
                                 output_name="clipped_buffer",
                                 context={"extent":{'xmin': -77.50941999999998,'ymin': 38.389560000000074,'xmax': -76.50941999999998,'ymax': 39.389560000000074,"spatialReference":{"wkid":102100,"latestWkid":3857}}})

    """
    kwargs = locals()
    tool_name = "ClipLayer"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value
    if output_name is None:
        output_service_name = 'Clip_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Overlay Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "clip_layer": (_FeatureSet, "clipLayer"),
        "outputType" : (str, 'outputType'),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "output"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return


def overlay_data(input_layer, 
                 overlay_layer, 
                 overlay_type="intersect", 
                 output_name=None, 
                 gis=None, 
                 include_overlaps=True,
                 context=None,
                 future=False):
    """
    .. image:: _static/images//overlay_layers/overlay_layers.png

    .. |Intersect| image:: _static/images/overlay_layers/Intersect.png
    .. |Union| image:: _static/images/overlay_layers/Union.png
    .. |Erase| image:: _static/images/overlay_layers/Erase.png
    .. |identity| image:: _static/images/overlay_layers/identity.png
    .. |symm| image:: _static/images/overlay_layers/Symmetric_Difference.png

    The ``overlay_data`` task combines two or more layers into one single layer. 
    You can think of overlay as peering through a stack of maps and creating a single 
    map containing all the information found in the stack. Overlay is used to answer 
    one of the most basic questions of geography: What is on top of what? 
    The following are examples:

        * What parcels are within the 100-year floodplain? ("Within" is another way of saying "on top of.")
        * What land use is within what soil type?
        * What mines are within abandoned military bases?

    .. note::
        Only available at ArcGIS Enterprise 10.6.1 and later.

    ======================  ===============================================================================
    **Argument**            **Description**
    ----------------------  -------------------------------------------------------------------------------
    input_layer             Required layer. The point, line, or polygon features that will be overlaid with the ``overlay_layer`` features. See :ref:`Feature Input<FeatureInput>`.
    ----------------------  -------------------------------------------------------------------------------
    overlay_layer           Required layer. The features that will be overlaid with the ``input_layer`` features.
    ----------------------  -------------------------------------------------------------------------------
    overlay_type            optional string. The type of overlay to be performed.

                            Choice list: ['intersect', 'erase', 'union', 'identity', 'symmetricaldifference']

                            +------------------------------------+-----------------------------------------------------------------------------------+
                            | |Intersect| ``intersect``          | Computes a geometric intersection of the input layers.                            |
                            |                                    | Features or portions of features that overlap in both the ``input_layer``         |
                            |                                    | and ``overlay_layer`` layers will be written to the output layer.                 |
                            |                                    |                                                                                   |
                            |                                    | * Point — Point, Line, Polygon                                                    |
                            |                                    | * Line — Point, Line, Polygon                                                     |
                            |                                    | * Polygon— Point, Line, Polygon                                                   |
                            +------------------------------------+-----------------------------------------------------------------------------------+
                            | |Erase| ``erase``                  | Only those features or portions of features in the ``overlay_layer``              |
                            |                                    | that are not within the features in the ``input_layer`` layer are                 |
                            |                                    | written to the output.                                                            |
                            |                                    |                                                                                   | 
                            |                                    | * Point — Point                                                                   |
                            |                                    | * Line — Line                                                                     |
                            |                                    | * Polygon — Polygon                                                               |
                            +------------------------------------+-----------------------------------------------------------------------------------+
                            | |Union| ``union``                  | Computes a geometric union of the ``input_layer`` and ``overlay_layer``.          |
                            |                                    | All features and their attributes will be written to the layer.                   |
                            |                                    |                                                                                   |
                            |                                    | * Polygon — Polygon                                                               |
                            +------------------------------------+-----------------------------------------------------------------------------------+
                            | |identity| ``identity``            | Computes a geometric intersection of the input features and                       |
                            |                                    | identity features. Features or portions of features that overlap in both          |
                            |                                    | ``input_layer`` and ``overlay_layer`` will be written to the output layer.        |
                            |                                    |                                                                                   |
                            |                                    |  * Point — Point, Polygon                                                         |
                            |                                    |  * Line — Line, Polygon                                                           |
                            |                                    |  * Polygon— Polygon                                                               |
                            +------------------------------------+-----------------------------------------------------------------------------------+
                            | |symm| ``symmetricaldifference``   | Features or portions of features in the ``input_layer``                           |
                            |                                    | and ``overlay_layer`` that do not overlap will be written to the output layer.    | 
                            |                                    |                                                                                   |
                            |                                    | * Point — Point                                                                   |
                            |                                    | * Line — Line                                                                     |
                            |                                    | * Polygon— Polygon                                                                |
                            +------------------------------------+-----------------------------------------------------------------------------------+
    ----------------------  -------------------------------------------------------------------------------
    include_overlaps        Optional boolean. Determines whether input features in the same dataset contain any overlapping features. 
                            This option should only be modified if you're not interested in self-intersection between 
                            features for the input layer and self-intersection between features for the overlay layer. 
                            Setting this value to false will improve performance. This parameter is only used when 
                            ``include_overlaps`` is Intersect with 10.6 and 10.6.1.

                            The default value is 'True'.
    ----------------------  -------------------------------------------------------------------------------
    output_name             Optional string. The task will create a feature service of the results. You define the name of the service.
    ----------------------  -------------------------------------------------------------------------------
    gis                     Optional GIS. The GIS object where the analysis will take place.
    ----------------------  -------------------------------------------------------------------------------
    context                 Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                            #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                            #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                            #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                            #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    ----------------------  -------------------------------------------------------------------------------
    future                  Optional boolean. If 'True', a GPJob is returned instead of
                            results. The GPJob can be queried on the status of the execution.

                            The default value is 'False'.
    ======================  ===============================================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To find the intersecting areas between watersheds and grazing land in Missouri.

            overlay_result = manage_data.overlay_data(input_layer=grazing_land, 
                                          overlay_layer=watersheds_layer, 
                                          overlay_type="Intersect", 
                                          output_name="Watershed_intersections")   
    """
    kwargs = locals()
    tool_name = "OverlayLayers"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json",
        "outputType" : "Input",
        'tolerance' : 0,
        'snapToInput' : 'false'
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Overlay_Layers_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Overlay Layers')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "overlay_layer": (_FeatureSet, "overlayLayer"),
        "outputType" : (str, 'outputType'),
        "overlay_type" : (str, "overlayType"),
        "output_name": (str, "OutputName"),
        "context": (str, "context"),
        'tolerance' : (int, 'tolerance'),
        "output": (_FeatureSet, "output"),
        'snapToInput' : (str, 'snapToInput')
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return


def append_data(input_layer, append_layer, field_mapping=None, gis=None, future=False):
    """
    Only available at ArcGIS Enterprise 10.6.1 and later.

    The Append Data task appends tabular, point, line, or polygon data to an existing layer.
    The input layer must be a hosted feature layer. The tool will add the appended data as
    rows to the input layer. No new output layer is created.

    ================  ===============================================================
    **Argument**      **Description**
    ----------------  ---------------------------------------------------------------
    input_layer       required FeatureLayer , The table, point, line or polygon features.
    ----------------  ---------------------------------------------------------------
    append_layer      required FeatureLayer. The table, point, line, or polygon features
                      to be appended to the input_layer. To append geometry, the
                      append_layer must have the same geometry type as the
                      input_layer. If the geometry types are not the same, the
                      append_layer geometry will be removed and all other matching
                      fields will be appended. The geometry of the input_layer will
                      always be maintained.
    ----------------  ---------------------------------------------------------------
    field_mapping     Defines how the fields in append_layer are appended to the
                      input_layer.

                      The following are set by default:

                        - All append_layer fields that match input_layer schema will be appended.
                        - Fields that exist in the input_layer and not in the append_layer will be appended with null values.
                        - Fields that exist in the append_layer and not in the input_layer will not be appended.

                      Optionally choose how input_layer fields will be appended from the following:

                      - AppendField - Matches the input_layer field with an append_layer field of a different name. Field types must match.
                      - Expression - Calculates values for the resulting field. Values are calculated using Arcade expressions. To assign null values, use 'null'.
    ----------------  ---------------------------------------------------------------
    gis               optional GIS, the GIS on which this tool runs. If not
                      specified, the active GIS is used.
    ----------------  ---------------------------------------------------------------
    future            optional Boolean. If True, a GPJob is returned instead of
                      results. The GPJob can be queried on the status of the execution.
    ================  ===============================================================

    :returns: boolean

    """
    kwargs = locals()
    tool_name = "AppendData"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "append_layer": (_FeatureSet, "appendLayer"),
        "field_mapping" : (str, "fieldMapping"),
        "context": (str, "context")
    }
    return_values = [
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=None)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return True
    except:
        raise

    return False


def calculate_fields(input_layer,
                     field_name,
                     data_type,
                     expression,
                     track_aware=False,
                     track_fields=None,
                     time_boundary_split=None,
                     time_split_unit=None,
                     time_reference=None,
                     output_name=None,
                     gis=None,
                     context=None,
                     future=False
                     ):
    """

    .. image:: _static/images/calculate_field/calculate_field.png 

    The ``calculate_fields`` task works with a layer to create and populate a
    new field. The output is a new feature layer, that is the same as the
    input features, with the additional field added.

    =================================================   ===============================================================
    **Argument**                                        **Description**
    -------------------------------------------------   ---------------------------------------------------------------
    input_layer                                         Required layer. The input features that will have a field added and calculated.
                                                        See :ref:`Feature Input<FeatureInput>`.
    -------------------------------------------------   ---------------------------------------------------------------
    field_name                                          Required string. A string representing the name of the new
                                                        field. If the name already exists in the dataset, then a
                                                        numeric value will be appended to the field name.
    -------------------------------------------------   ---------------------------------------------------------------
    data_type                                           Required string. The type for the new field.
    
                                                        Choice list: ['Date', 'Double', 'Integer', 'String'`].
    -------------------------------------------------   ---------------------------------------------------------------
    expression                                          Required string. An Arcade expression used to calculate the new
                                                        field values. You can use any of the Date, Logical,
                                                        Mathematical or Text function available with Arcade.
    -------------------------------------------------   ---------------------------------------------------------------
    track_aware                                         Optional boolean. Boolean value denoting if the expression is
                                                        track aware.

                                                        The default value is 'False'.
    -------------------------------------------------   ---------------------------------------------------------------
    track_fields (Required if trackAware is True)       Optional string. The fields used to identify distinct tracks.
                                                        There can be multiple track_fields. track_fields are only
                                                        required when ``track_aware`` is True.
    -------------------------------------------------   ---------------------------------------------------------------
    time_boundary_split                                 Optional integer. A time boundary allows your to analyze values within a defined time span. 
                                                        For example, if you use a time boundary of 1 day, starting on January 1st, 1980 tracks will 
                                                        be analyzed 1 day at a time. The time boundary parameter was introduced in ArcGIS Enterprise 10.7.

                                                        The time boundary parameters are only applicable if the analysis is ``track_aware``.
                                                        The ``time_boundary_split`` parameter defines the scale of the time boundary. 
                                                        In the case above, this would be 1. See the portal documentation for this tool to learn more.
    -------------------------------------------------   ---------------------------------------------------------------
    time_split_unit                                     Optional string.  The unit to detect an incident is `time_boundary_split` is used.

                                                        Choice list: ['Years', 'Months', 'Weeks', 'Days', 'Hours', 'Minutes', 'Seconds', 'Milliseconds'].
    -------------------------------------------------   ---------------------------------------------------------------
    time_reference                                      Optional datetime.datetime. The starting date/time where analysis will
                                                        begin from.
    -------------------------------------------------   ---------------------------------------------------------------
    output_name                                         Optional string, The task will create a feature service of the
                                                        results. You define the name of the service.
    -------------------------------------------------   ---------------------------------------------------------------
    gis                                                 Optional GIS, the GIS on which this tool runs. If not
                                                        specified, the active GIS is used.
    -------------------------------------------------   ---------------------------------------------------------------
    context                                             Optional dict. The context parameter contains additional settings that affect task execution. For this task, there are four settings:

                                                        #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                                        #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                                        #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                                        #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
    =================================================   ===============================================================
    

    :returns: feature layer collection

    .. code-block:: python

            # Usage Example: To find maximum of the two attributes of an input layer.

            result = calculate_fields(input_layer=lyr,
                                      field_name="avg",
                                      data_type="Double",
                                      expression='max($feature["InputValue"],$feature["Value2"])')
    """
    kwargs = locals()
    tool_name = "CalculateField"
    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url
    params = {
        "f" : "json"
    }
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Calculate_Fields_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')

    output_service = _create_output_service(gis, output_name, output_service_name, 'Calculate Fields')

    params['output_name'] = _json.dumps({
        "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
        "itemProperties": {"itemId" : output_service.itemid}})

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "field_name" : (str, "fieldName"),
        "data_type" : (str, "dataType"),
        "expression" : (str, "expression"),
        "track_aware" : (bool, "trackAware"),
        "track_fields" : (str, "trackFields"),
        "time_boundary_split" : (int, "timeBoundarySplit"),
        "time_split_unit" : (str, "timeBoundarySplitUnit"),
        "time_reference" : (_datetime.datetime, "timeBoundaryReference"),
        "output_name": (str, "outputName"),
        "output": (_FeatureSet, "output"),
        "context": (str, "context")
    }
    return_values = [
        {"name": "output", "display_name": "Output Features", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, tool_name, params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        output_service.delete()
        raise

    return

def copy_to_data_store(
    input_layer,
    output_name=None,
    gis=None,
    context=None,
    future=False):
    """
    .. image:: _static/images/copy_to_data_store/copy_to_data_store.png 

    The ``copy_to_data_store`` task takes an input layer and copies it to a data store. 
    Data is copied to ArcGIS Data Store, configured as either a relational or 
    spatiotemporal big data store.
    
    For example

        * Copy a collection of .csv files in a big data file share to the spatiotemporal data store for visualization.
        * Copy the features in the current map extent that are stored in the spatiotemporal data store to the relational data store.

    This tool will take an input layer and copy it to a data store. Data will be copied to the ArcGIS Data Store 
    and will be stored in your relational or spatiotemporal data store.

    For example, you could copy features that are stored in a big data file share to a relational data store 
    and specify that only features within the current map extent will be copied. This would create a hosted 
    feature service with only those features that were within the specified map extent.

    ==========================   ===============================================================
    **Argument**                 **Description**
    --------------------------   ---------------------------------------------------------------
    input_layer                  Required layer. The table, point, line, or polygon features that will be copied. See :ref:`Feature Input<FeatureInput>`.
    --------------------------   ---------------------------------------------------------------
    output_name                  Optional string. The task will create a feature service of the results. You define the name of the service.
    --------------------------   --------------------------------------------------------------- 
    gis                          Optional GIS. The GIS on which this tool runs. If not specified, the active GIS is used.
    --------------------------   --------------------------------------------------------------- 
    context                      Optional string. The context parameter contains additional settings that affect task execution. For this task, there are five settings:

                                 #. Extent (``extent``) - A bounding box that defines the analysis area. Only those features that intersect the bounding box will be analyzed.
                                 #. Processing spatial reference (``processSR``) - The features will be projected into this coordinate system for analysis.
                                 #. Output spatial reference (``outSR``) - The features will be projected into this coordinate system after the analysis to be saved. The output spatial reference for the spatiotemporal big data store is always WGS84.
                                 #. Data store (``dataStore``) - Results will be saved to the specified data store. The default is the spatiotemporal big data store.
                                 #. Default aggregation styles (``defaultAggregationStyles``) - If set to 'True', results will have square, hexagon, and triangle aggregation styles enabled on results map services.
    --------------------------   --------------------------------------------------------------- 
     future                      Optional boolean. If 'True', the result comes back as a GPJob.

                                 The default value is 'False'.
    ==========================   ===============================================================

    :returns: result_layer : Output Features as feature layer item.

    .. code-block:: python

            # Usage Example: To copy input layer to a data store. 
            copy_result = copy_to_data_store(input_layer=earthquakes,
                                             output_name="copy earthquakes data")
    """
    kwargs = locals()

    gis = _arcgis.env.active_gis if gis is None else gis
    url = gis.properties.helperServices.geoanalytics.url

    params = {}
    for key, value in kwargs.items():
        if value is not None:
            params[key] = value

    if output_name is None:
        output_service_name = 'Data Store Copy_' + _id_generator()
        output_name = output_service_name.replace(' ', '_')
    else:
        output_service_name = output_name.replace(' ', '_')
    _set_context(params)
    if 'context' in params and \
       params['context'].lower().find('bigdatafileshares') > -1:
        params['output_name'] = output_name
        output_service = True
    else:
        output_service = _create_output_service(gis, output_name, output_service_name, 'Copy To Data Store')
        params['output_name'] = _json.dumps({
            "serviceProperties": {"name" : output_name, "serviceUrl" : output_service.url},
            "itemProperties": {"itemId" : output_service.itemid}})

    #_set_context(params)

    if context is not None:
        params["context"] = context
    else:
        _set_context(params)

    param_db = {
        "input_layer": (_FeatureSet, "inputLayer"),
        "output_name": (str, "outputName"),
        "context": (str, "context"),
        "output": (_FeatureSet, "Output Layer"),
    }
    return_values = [
        {"name": "output", "display_name": "Output Layer", "type": _FeatureSet},
    ]
    try:
        if future:
            gpjob = _execute_gp_tool(gis, "CopyToDataStore", params, param_db, return_values, _use_async, url, True, future=future)
            return GAJob(gpjob=gpjob, return_service=output_service)
        _execute_gp_tool(gis, "CopyToDataStore", params, param_db, return_values, _use_async, url, True, future=future)
        return output_service
    except:
        if isinstance(output_service, bool):
            return False
        output_service.delete()
        raise

copy_to_data_store.__annotations__ = {
    'output_name': str}





